/* interpolation/gsl_polynomial.c
 * 
 * Copyright (C) 2001 DAN, HO-JIN
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_interp.h>
#include "gsl_polynomial.h"
#include "interp_poly.h"

gsl_polynomial* gsl_polynomial_alloc (const gsl_interp_type * T, size_t size)
{
  gsl_polynomial *polynomial =
    (gsl_polynomial *) malloc (sizeof (gsl_polynomial));

  if (polynomial == NULL)
    {
      GSL_ERROR_NULL ("failed to allocate space for polynomial struct",
		      GSL_ENOMEM);
    }

  polynomial->interp = gsl_interp_alloc (T, size);

  if (polynomial->interp == NULL)
    {
      free (polynomial);
      GSL_ERROR_NULL ("failed to allocate space for interp", GSL_ENOMEM);
    };

  polynomial->x = (double *) malloc (size * sizeof (double));

  if (polynomial->x == NULL)
    {
      gsl_interp_free (polynomial->interp);
      free (polynomial);
      GSL_ERROR_NULL ("failed to allocate space for x", GSL_ENOMEM);
    }

  polynomial->y = (double *) malloc (size * sizeof (double));

  if (polynomial->y == NULL)
    {
      free (polynomial->x);
      gsl_interp_free (polynomial->interp);
      free (polynomial);
      GSL_ERROR_NULL ("failed to allocate space for y", GSL_ENOMEM);
    }

  polynomial->size = size;

  return polynomial;
}

int
gsl_polynomial_init (gsl_polynomial * polynomial, const double x_array[],
		     const double y_array[], size_t size)
{
  if (size != polynomial->size)
    {
      GSL_ERROR ("data must match size of polynomial object", GSL_EINVAL);
    }

  memcpy (polynomial->x, x_array, size * sizeof (double));
  memcpy (polynomial->y, y_array, size * sizeof (double));

  {
    int status = gsl_interp_init (polynomial->interp, x_array, y_array, size);
    return status;
  }
}

void
gsl_polynomial_free (gsl_polynomial * polynomial)
{
  gsl_interp_free (polynomial->interp);
  free (polynomial->x);
  free (polynomial->y);
  free (polynomial);
}

int
gsl_polynomial_eval_e (const gsl_polynomial * polynomial, double x, double *y)
{
  return gsl_interp_eval_e (polynomial->interp,
			    polynomial->x, polynomial->y, x, 0, y);
}

double
gsl_polynomial_eval (const gsl_polynomial * polynomial, double x)
{
  return gsl_interp_eval (polynomial->interp,
			  polynomial->x, polynomial->y, x, 0);
}


int
gsl_polynomial_eval_deriv_e (const gsl_polynomial * polynomial,
			     double x, double *dydx)
{
  return gsl_interp_eval_deriv_e (polynomial->interp,
				  polynomial->x, polynomial->y, x, 0, dydx);
}

double
gsl_polynomial_eval_deriv (const gsl_polynomial * polynomial, double x)
{
  return gsl_interp_eval_deriv (polynomial->interp,
				polynomial->x, polynomial->y, x, 0);
}


int
gsl_polynomial_eval_deriv2_e (const gsl_polynomial * polynomial,
			      double x, double *d2)
{
  return gsl_interp_eval_deriv2_e (polynomial->interp,
				   polynomial->x, polynomial->y, x, 0, d2);
}

double
gsl_polynomial_eval_deriv2 (const gsl_polynomial * polynomial, double x)
{
  return gsl_interp_eval_deriv2 (polynomial->interp,
				 polynomial->x, polynomial->y, x, 0);
}


int
gsl_polynomial_eval_integ_e (const gsl_polynomial * polynomial,
			     double a, double b, double *result)
{
  return gsl_interp_eval_integ_e (polynomial->interp,
				  polynomial->x, polynomial->y,
				  a, b, 0, result);
}


double
gsl_polynomial_eval_integ (const gsl_polynomial * polynomial,
			   double a, double b)
{
  return gsl_interp_eval_integ (polynomial->interp,
				polynomial->x, polynomial->y, a, b, 0);
}
