/* interpolation/interp_poly.c
 * 
 * Copyright (C) 2001 DAN, HO-JIN
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <math.h>
#include <stdio.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_interp.h>
#include <gsl/gsl_vector.h>
#include "interp_poly.h"

/* for ctags ref. 
 * gsl_interp_alloc
 * gsl_spline_alloc
 * gsl_interp
 * cspline_state_t
 * gsl_interp_cspline
 * gsl_spline
 */

typedef struct
{
  double *d;
}
polynomial_state_t;

void *
polynomial_alloc (size_t size)
{
  polynomial_state_t *state =
    (polynomial_state_t *) malloc (sizeof (polynomial_state_t));
  if (state == 0)
    {
      GSL_ERROR_NULL ("failed to allocate space for polynomial_state_t",
		      GSL_ENOMEM);
    }
  state->d = (double *) malloc (sizeof (double) * size);
  if (state->d == 0)
    {
      free (state);
      GSL_ERROR_NULL ("failed to allocate space for polynomial_state_t.d",
		      GSL_ENOMEM);
    }
  return state;
}

int
polynomial_init (void *vstate,
		 const double xa[], const double ya[], size_t size)
{
  polynomial_state_t *state = (polynomial_state_t *) vstate;
  size_t i, j;
  /* newton's divide difference */
  state->d[0] = ya[0];
  for (j = size - 1; j >= 1; j--)
    {
      state->d[j] = (ya[j] - ya[j - 1]) / (xa[j] - xa[j - 1]);
    }
  for (i = 2; i < size; i++)
    {
      for (j = size - 1; j >= i; j--)
	{
	  state->d[j] = (state->d[j] - state->d[j - 1]) / (xa[j] - xa[j - i]);
	}
    }
  /*
  for (i = 0; i < size; i++)
    printf ("[%d]%e\n", i, state->d[i]);
    */
  return GSL_SUCCESS;
}

int
polynomial_eval (const void *vstate,
		 const double xa[], const double ya[], size_t size, double x,
		 gsl_interp_accel * acc, double *y)
{
  polynomial_state_t *state = (polynomial_state_t *) vstate;
  int i;
  *y = state->d[size - 1];
  for (i = size - 2; i >= 0; i--)
    {
      *y = state->d[i] + (x - xa[i]) * (*y);
    }
  return GSL_SUCCESS;
}

int
polynomial_taylor (const void *vstate, const double x[], const double f[],
		   size_t size, double x_0, double *coef)
{
  polynomial_state_t *state = (polynomial_state_t *) vstate;
  double *w = 0;
  int i, j;
  w = (double *) calloc (sizeof (double), size);
  if (w == 0)
    {
      GSL_ERROR_NULL ("failed to allocate workspace in polynomial_taylor",
		      GSL_ENOMEM);
    }
  /* coef[] must be cleaned before */
  {
    gsl_vector_view v = gsl_vector_view_array (coef, size);
    gsl_vector_set_zero (&v.vector);
  }
  w[size - 1] = 1.;
  coef[0] = state->d[0];
  for (i = size - 2; i >= 0; i--)
    {
      w[i] = -w[i + 1] * (x[size - 2 - i] - x_0);
      for (j = i + 1; j <= size - 2; j++)
	{
	  w[j] = w[j] - w[j + 1] * (x[size - 2 - i] - x_0);
	}
      for (j = i; j < size; j++)
	{
	  coef[j - i] += w[j] * state->d[size - i - 1];
	}
    }
  /* 
  for (j=0; j<size; j++) {
    printf("c[%i]=%f, ", j, coef[j]);
  }
  putchar('\n');
  */
  free (w);
  return GSL_SUCCESS;
}

int
polynomial_deriv (const void *vstate,
		  const double xa[], const double ya[], size_t size, double x,
		  gsl_interp_accel * acc, double *y)
{
  double *coef = 0;
  coef = (double *) calloc (sizeof (double), size);
  if (coef == 0)
    {
      GSL_ERROR_NULL ("failed to allocate workspace in polynomial_deriv",
		      GSL_ENOMEM);
    }
  polynomial_taylor (vstate, xa, ya, size, x, coef);
  *y = coef[1];
  free (coef);
  return GSL_SUCCESS;
}

int
polynomial_deriv2 (const void *vstate,
		   const double xa[], const double ya[], size_t size,
		   double x, gsl_interp_accel * acc, double *y)
{
  double *coef = 0;
  coef = (double *) calloc (sizeof (double), size);
  if (coef == 0)
    {
      GSL_ERROR_NULL ("failed to allocate workspace in polynomial_deriv2",
		      GSL_ENOMEM);
    }
  polynomial_taylor (vstate, xa, ya, size, x, coef);
  *y = 2. * coef[2];
  free (coef);
  return GSL_SUCCESS;
}

int
polynomial_integ (const void *vstate, const double xa[], const double ya[],
		  size_t size, gsl_interp_accel * acc, double a, double b,
		  double *result)
{
  double *coef = 0;
  int i;
  coef = (double *) calloc (sizeof (double), size);
  if (coef == 0)
    {
      GSL_ERROR_NULL ("failed to allocate workspace in polynomial_deriv2",
		      GSL_ENOMEM);
    }
  polynomial_taylor (vstate, xa, ya, size, 0., coef);
  *result = coef[0] * (b - a);
  for (i = 1; i < size; i++)
    {
      *result += coef[i] / (1. + (double) i) * (pow (b, i+1) - pow (a, i+1));
    }
  free (coef);
  return GSL_SUCCESS;
}

void
polynomial_free (void *vstate)
{
  polynomial_state_t *state = (polynomial_state_t*)vstate;
  free (state->d);
  free (state);
}

static const gsl_interp_type polynomial_type = {
  "polynomial",
  3,
  &polynomial_alloc,
  &polynomial_init,
  &polynomial_eval,
  &polynomial_deriv,
  &polynomial_deriv2,
  &polynomial_integ,
  &polynomial_free,
};

const gsl_interp_type *gsl_interp_polynomial = &polynomial_type;
